@@ -1,11 +1,11 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header$
+ * $Header: /cvsroot/pgpool/pgpool-II/pool_timestamp.c,v 1.20 2011/09/05 09:49:37 kitagawa Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2013	PgPool Global Development Group
+ * Copyright (c) 2003-2011	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -29,13 +29,13 @@
 #include "pool_select_walker.h"
 #include "pool_config.h"
 #include "parser/parsenodes.h"
-#include "parser/gramparse.h"
+#include "parser/parser.h"
 #include "parser/pool_memory.h"
 
-#define Assert(x)
 
 typedef struct {
-	char	*attrname;
+	char	*attrname;	/* attribute name */
+	char	*adsrc;		/* default value expression */
 	int		 use_timestamp;
 } TSAttr;
 
@@ -48,10 +48,10 @@ typedef struct {
 	A_Const					*ts_const;
 	POOL_CONNECTION_POOL	*backend;
 	char					*relname;
-	int						 num_params;		/* num of original params (for Parse) */
-	bool		 			 rewrite_to_params;	/* true if rewritten to param insread of const */
-	bool		 			 rewrite;			/* has rewritten? */
-	List					*params;			/* list of additional params */
+	int						 num_params;	/* num of original params (for Parse) */
+	bool		 			 rewrite_to_params;
+	bool		 			 rewrite;		/* has rewritten? */
+	List					*params;		/* list of additional params */
 } TSRewriteContext;
 
 static void *ts_register_func(POOL_SELECT_RESULT *res);
@@ -64,6 +64,7 @@ static bool rewrite_timestamp_insert(InsertStmt *i_stmt, TSRewriteContext *ctx);
 static bool rewrite_timestamp_update(UpdateStmt *u_stmt, TSRewriteContext *ctx);
 static char *get_current_timestamp(POOL_CONNECTION_POOL *backend);
 static Node *makeTsExpr(TSRewriteContext *ctx);
+static A_Const *makeStringConstFromQuery(POOL_CONNECTION_POOL *backend, char *expression);
 bool raw_expression_tree_walker(Node *node, bool (*walker) (), void *context);
 
 #define		MAX_RELCACHE 32
@@ -73,6 +74,9 @@ POOL_RELCACHE	*ts_relcache;
 static void *
 ts_register_func(POOL_SELECT_RESULT *res)
 {
+/* Number of result columns included in res */
+#define NUM_COLS		3
+
 	TSRel	*rel;
 	int		 i;
 
@@ -83,10 +87,22 @@ ts_register_func(POOL_SELECT_RESULT *res)
 
 	for (i = 0; i < res->numrows; i++)
 	{
-		rel->attr[i].attrname = strdup(res->data[i * 2]);
-		rel->attr[i].use_timestamp = *(res->data[i * 2 + 1]) == 't';
-		pool_debug("attrname %s use_timestamp = %d",
-			rel->attr[i].attrname, rel->attr[i].use_timestamp);
+		int index = 0;
+
+		rel->attr[i].attrname = strdup(res->data[i * NUM_COLS + index]);
+		index++;
+
+		if (res->data[i * NUM_COLS + index])
+			rel->attr[i].adsrc = strdup(res->data[i * NUM_COLS + index]);
+		else
+			rel->attr[i].adsrc = NULL;
+
+		index++;
+
+		rel->attr[i].use_timestamp = *(res->data[i * NUM_COLS + index]) == 't';
+		pool_debug("attrname %s adsrc %s use_timestamp = %d",
+				   rel->attr[i].attrname, (rel->attr[i].adsrc? rel->attr[i].adsrc:"NULL"),
+										   rel->attr[i].use_timestamp);
 	}
 
 	rel->relnatts = res->numrows;
@@ -97,17 +113,20 @@ ts_register_func(POOL_SELECT_RESULT *res)
 static void *
 ts_unregister_func(void *data)
 {
-    TSRel   *rel = (TSRel *) data;
-    if (rel)
-        free(rel);
-    return (void *)0;
+	TSRel	*rel = (TSRel *) data;
+
+	if (rel == NULL)
+		return NULL;
+
+	free(rel);
+	return rel;
 }
 
 
 static TSRel*
 relcache_lookup(TSRewriteContext *ctx)
 {
-#define ATTRDEFQUERY "SELECT attname, coalesce((d.adsrc = 'now()' OR d.adsrc LIKE '%%''now''::text%%')" \
+#define ATTRDEFQUERY "SELECT attname, d.adsrc, coalesce((d.adsrc LIKE '%%now()%%' OR d.adsrc LIKE '%%''now''::text%%')" \
 	" AND (a.atttypid = 'timestamp'::regtype::oid OR" \
 	" a.atttypid = 'timestamp with time zone'::regtype::oid OR" \
 	" a.atttypid = 'date'::regtype::oid OR" \
@@ -119,7 +138,7 @@ relcache_lookup(TSRewriteContext *ctx)
 	" WHERE c.oid = a.attrelid AND a.attnum >= 1 AND a.attisdropped = 'f' AND c.relname = '%s'" \
 	" ORDER BY a.attnum"
 
-#define ATTRDEFQUERY2 "SELECT attname, coalesce((d.adsrc = 'now()' OR d.adsrc LIKE '%%''now''::text%%')" \
+#define ATTRDEFQUERY2 "SELECT attname, d.adsrc, coalesce((d.adsrc LIKE '%%now()%%' OR d.adsrc LIKE '%%''now''::text%%')" \
 	" AND (a.atttypid = 'timestamp'::regtype::oid OR" \
 	" a.atttypid = 'timestamp with time zone'::regtype::oid OR" \
 	" a.atttypid = 'date'::regtype::oid OR" \
@@ -156,6 +175,7 @@ relcache_lookup(TSRewriteContext *ctx)
 	return (TSRel *) pool_search_relcache(ts_relcache, ctx->backend, ctx->relname);
 }
 
+
 static Node *
 makeTsExpr(TSRewriteContext *ctx)
 {
@@ -217,7 +237,7 @@ isSystemTypeCast(Node *node, const char *name)
 		return false;
 
 	typecast = (TypeCast *) node;
-	return isSystemType((Node *) typecast->typename, name);
+	return isSystemType((Node *) typecast->typeName, name);
 }
 
 /*
@@ -236,12 +256,15 @@ rewrite_timestamp_walker(Node *node, void *context)
 		/* `now()' FuncCall */
 		FuncCall	*fcall = (FuncCall *) node;
 
-		if (list_length(fcall->funcname) == 1 &&
-			strcmp("now", strVal(linitial(fcall->funcname))) == 0)
+		if ((list_length(fcall->funcname) == 1 &&
+			 strcmp("now", strVal(linitial(fcall->funcname))) == 0) ||
+			(list_length(fcall->funcname) == 2 &&
+			 strcmp("pg_catalog", strVal(linitial(fcall->funcname))) == 0 &&
+			 strcmp("now", strVal(lsecond(fcall->funcname))) == 0))
 		{
 			TypeCast	*tc = makeNode(TypeCast);
 			tc->arg = makeTsExpr(ctx);
-			tc->typename = SystemTypeName("text");
+			tc->typeName = SystemTypeName("text");
 
 			fcall->funcname = SystemFuncName("timestamptz");
 			fcall->args = list_make1(tc);
@@ -253,11 +276,11 @@ rewrite_timestamp_walker(Node *node, void *context)
 		/* CURRENT_DATE, CURRENT_TIME, LOCALTIMESTAMP, LOCALTIME etc.*/
 		TypeCast	*tc = (TypeCast *) node;
 
-		if ((isSystemType((Node *) tc->typename, "date") ||
-			 isSystemType((Node *) tc->typename, "timestamp") ||
-			 isSystemType((Node *) tc->typename, "timestamptz") ||
-			 isSystemType((Node *) tc->typename, "time") ||
-			 isSystemType((Node *) tc->typename, "timetz")))
+		if ((isSystemType((Node *) tc->typeName, "date") ||
+			 isSystemType((Node *) tc->typeName, "timestamp") ||
+			 isSystemType((Node *) tc->typeName, "timestamptz") ||
+			 isSystemType((Node *) tc->typeName, "time") ||
+			 isSystemType((Node *) tc->typeName, "timetz")))
 		{
 			/* rewrite `'now'::timestamp' and `'now'::text::timestamp' both */
 			if (isSystemTypeCast(tc->arg, "text"))
@@ -291,7 +314,7 @@ get_current_timestamp(POOL_CONNECTION_POOL *backend)
 {
 	POOL_SELECT_RESULT *res;
 	POOL_STATUS		 status;
-	static char		timestamp[64];
+	static char		timestamp[32];
 
 	status = do_query(MASTER(backend), "SELECT now()", &res, MAJOR(backend));
 	if (status != POOL_CONTINUE)
@@ -344,7 +367,11 @@ rewrite_timestamp_insert(InsertStmt *i_stmt, TSRewriteContext *ctx)
 			if (relcache->attr[i].use_timestamp)
 			{
 				rewrite = true;
-				values = lappend(values, makeTsExpr(ctx));
+				if (ctx->rewrite_to_params)
+					values = lappend(values, makeTsExpr(ctx));
+				else
+					values = lappend(values,
+									 makeStringConstFromQuery(ctx->backend, relcache->attr[i].adsrc));
 			}
 			else
 				values = lappend(values, makeNode(SetToDefault));
@@ -407,7 +434,10 @@ rewrite_timestamp_insert(InsertStmt *i_stmt, TSRewriteContext *ctx)
 					if (relcache->attr[i].use_timestamp == true && IsA(lfirst(lc_val), SetToDefault))
 					{
 						rewrite = true;
-						lfirst(lc_val) = makeTsExpr(ctx);
+						if (ctx->rewrite_to_params)
+							lfirst(lc_val) = makeTsExpr(ctx);
+						else
+							lfirst(lc_val) = makeStringConstFromQuery(ctx->backend, relcache->attr[i].adsrc);
 					}
 					i++;
 				}
@@ -418,7 +448,11 @@ rewrite_timestamp_insert(InsertStmt *i_stmt, TSRewriteContext *ctx)
 					if (relcache->attr[i].use_timestamp == true)
 					{
 						rewrite = true;
-						values = lappend(values, makeTsExpr(ctx));
+						if (ctx->rewrite_to_params)
+							values = lappend(values, makeTsExpr(ctx));
+						else
+							values = lappend(values,
+											 makeStringConstFromQuery(ctx->backend, relcache->attr[i].adsrc));
 					}
 					else
 						values = lappend(values, makeNode(SetToDefault));
@@ -430,12 +464,15 @@ rewrite_timestamp_insert(InsertStmt *i_stmt, TSRewriteContext *ctx)
 			/*
 			 * INSERT INTO rel(col1, col2) VALUES (val, val2)
 			 *
-			 * if timestamp column does not given by column list
+			 * if timestamp column is not given by column list
 			 * add colname to column list and add timestamp to values list.
 			 */
 			int			append_columns = 0;
+			int			*append_columns_list;
 			ResTarget	*col;
 
+			append_columns_list = (int *)malloc(sizeof(int)*relcache->relnatts);
+
 			for (i = 0; i < relcache->relnatts; i++)
 			{
 				if (relcache->attr[i].use_timestamp == false)
@@ -458,7 +495,7 @@ rewrite_timestamp_insert(InsertStmt *i_stmt, TSRewriteContext *ctx)
 					col->indirection = NIL;
 					col->val = NULL;
 					i_stmt->cols = lappend(i_stmt->cols, col);
-					append_columns++;
+					append_columns_list[append_columns++] = i;
 				}
 			}
 
@@ -479,16 +516,24 @@ rewrite_timestamp_insert(InsertStmt *i_stmt, TSRewriteContext *ctx)
 					if (relcache->attr[i].use_timestamp == true && IsA(lfirst(lc_val), SetToDefault))
 					{
 						rewrite = true;
-						lfirst(lc_val) = makeTsExpr(ctx);
+						if (ctx->rewrite_to_params)
+							lfirst(lc_val) = makeTsExpr(ctx);
+						else
+							lfirst(lc_val) = makeStringConstFromQuery(ctx->backend, relcache->attr[i].adsrc);
 					}
 				}
 
 				/* add ts_const to values list */
 				for (i = 0; i < append_columns; i++)
 				{
-					values = lappend(values, makeTsExpr(ctx));
+					if (ctx->rewrite_to_params)
+						values = lappend(values, makeTsExpr(ctx));
+					else
+						values = lappend(values,
+										 makeStringConstFromQuery(ctx->backend, relcache->attr[append_columns_list[i]].adsrc));
 				}
 			}
+			free(append_columns_list);
 		}
 	}
 
@@ -539,7 +584,10 @@ rewrite_timestamp_update(UpdateStmt *u_stmt, TSRewriteContext *ctx)
 				{
 					if (relcache->attr[i].use_timestamp)
 					{
-						res->val = (Node *) makeTsExpr(ctx);
+						if (ctx->rewrite_to_params)
+							res->val = (Node *) makeTsExpr(ctx);
+						else
+							res->val = (Node *)makeStringConstFromQuery(ctx->backend, relcache->attr[i].adsrc);
 						rewrite = true;
 					}
 					break;
@@ -640,10 +688,6 @@ rewrite_timestamp(POOL_CONNECTION_POOL *backend, Node *node,
 	else
 		;
 
-	/* save number of parameters in original query */
-	if (message)
-		message->query_context->num_original_params = ctx.num_params;
-
 	if (!rewrite)
 		return NULL;
 
@@ -704,8 +748,7 @@ bind_rewrite_timestamp(POOL_CONNECTION_POOL *backend,
 	int32		 tmp4;
 	int			 i,
 				 ts_len,
-				 copy_len,
-				 num_org_params;
+				 copy_len;
 	const char	*copy_from;
 	char		*ts,
 				*copy_to,
@@ -728,11 +771,8 @@ bind_rewrite_timestamp(POOL_CONNECTION_POOL *backend,
 
 	ts_len = strlen(ts);
 
-	/* enlarge length for timestamp parameters */
-	*len += (ts_len + sizeof(int32)) * message->num_tsparams;
-	/* allocate extra memory for parameter formats */
-	num_org_params = message->query_context->num_original_params;
-	new_msg = copy_to = (char *) malloc(*len + sizeof(int16) * (message->num_tsparams + num_org_params));
+	*len += (strlen(ts) + sizeof(int32)) * message->num_tsparams;
+	new_msg = copy_to = (char *) malloc(*len + message->num_tsparams * sizeof(int16));
 	copy_from = orig_msg;
 
 	/* portal_name */
@@ -750,20 +790,12 @@ bind_rewrite_timestamp(POOL_CONNECTION_POOL *backend,
 	copy_len = sizeof(int16);
 	tmp2 = num_formats = ntohs(tmp2);
 
-	if (num_formats >= 1)
+	if (num_formats > 1)
 	{
-		/* one means the specified format code is applied all original parameters */
-		if (num_formats == 1)
-		{
-			*len += (num_org_params - 1) * sizeof(int16);
-			tmp2 += num_org_params - 1;
-		}
-
-		/* enlarge message length for timestamp parameter's formats */
+		/* enlarge message length */
 		*len += message->num_tsparams * sizeof(int16);
 		tmp2 += message->num_tsparams;
 	}
-
 	tmp2 = htons(tmp2);
 	memcpy(copy_to, &tmp2, copy_len);	/* copy number of format codes */
 	copy_to += copy_len; copy_from += copy_len;
@@ -773,17 +805,10 @@ bind_rewrite_timestamp(POOL_CONNECTION_POOL *backend,
 	memcpy(copy_to, copy_from, copy_len);		/* copy format codes */
 	copy_to += copy_len; copy_from += copy_len;
 
-	if (num_formats >= 1)
+	if (num_formats > 1)
 	{
-		/* copy the specified format code as numbers of original parameters */
-		if (num_formats == 1)
-		{
-			memcpy(copy_to, copy_from, (num_org_params - 1) * sizeof(int16));
-			copy_to += (num_org_params - 1) * sizeof(int16);
-		}
-
-		/* set additional format codes to zero(text) */
-		memset(copy_to, 0, message->num_tsparams * sizeof(int16));
+		/* set format codes to zero(text) */
+		memset(copy_to, 0, message->num_tsparams * 2);
 		copy_to += sizeof(int16) * message->num_tsparams;
 	}
 
@@ -832,6 +857,39 @@ bind_rewrite_timestamp(POOL_CONNECTION_POOL *backend,
 	return new_msg;
 }
 
+static A_Const *makeStringConstFromQuery(POOL_CONNECTION_POOL *backend, char *expression)
+{
+	A_Const *con;
+	POOL_SELECT_RESULT *res;
+	POOL_STATUS		 status;
+	char query[1024];
+	int len;
+	char *str;
+
+	snprintf(query, sizeof(query), "SELECT %s", expression);
+	status = do_query(MASTER(backend), query, &res, MAJOR(backend));
+	if (status != POOL_CONTINUE)
+	{
+		pool_error("makeStringConstFromQuery: do_query faild");
+		return NULL;
+	}
+
+	if (res->numrows != 1)
+	{
+		free_select_result(res);
+		return NULL;
+	}
+
+	len = strlen(res->data[0]) + 1;
+	str = palloc(len);
+	strcpy(str, res->data[0]);
+	free_select_result(res);
+
+	con = makeNode(A_Const);
+	con->val.type = T_String;
+	con->val.val.str = str;
+	return con;
+}
 
 /* from nodeFuncs.c start */
 
@@ -1070,7 +1128,7 @@ bool
 
 				if (walker(tc->arg, context))
 					return true;
-				if (walker(tc->typename, context))
+				if (walker(tc->typeName, context))
 					return true;
 			}
 			break;
@@ -1121,7 +1179,7 @@ bool
 			{
 				ColumnDef  *coldef = (ColumnDef *) node;
 
-				if (walker(coldef->typename, context))
+				if (walker(coldef->typeName, context))
 					return true;
 				if (walker(coldef->raw_default, context))
 					return true;
@@ -1136,7 +1194,7 @@ bool
 
 				if (walker(xs->expr, context))
 					return true;
-				if (walker(xs->typename, context))
+				if (walker(xs->typeName, context))
 					return true;
 			}
 			break;
